#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include "wzinfo.h"

//#define _DEBUG_K1

#define READ(fd,pointer,length) (read_retval=read((fd),(pointer),(length)))
#define _WZ_DBGVAR int read_retval;
#ifdef _DEBUG_K1
# define _WZ_READFILE printf("%s:%d - %s\n",__FILE__,__LINE__,strerror(errno)),((read_retval==0)?WZ_EOF:WZ_READFILE)
#else
# define _WZ_READFILE ((read_retval==0)?WZ_EOF:WZ_READFILE)
#endif

// This is pointer to current DataZeug .project
struct wzproject *project=NULL;

/*
 *  Delete WZ project
 */
void delete_wzproject(struct wzproject *wz) {
    unsigned int i;
    struct block *blk;
    
    if (wz!=NULL) {
        if (wz->blocks!=NULL) {
            for (i=0;i<wz->nblocks;i++) {
                blk=&(wz->blocks[i]);
                //free(blk->para);
                free(blk->name);
                free(blk->blob);
            };
            free(wz->blocks);
        };
        if (wz->pages!=NULL)
            free(wz->pages);
        free(wz);
    };
}

/*
 *  Load WZ project from file
 */
int load_wzproject(struct wzproject **wzinfo, int fd) {
    int retval;
    struct wzproject *opened;
    
    retval=_load_wzproject(&opened,fd);
    if (retval<0) {
        delete_wzproject(opened);
        *wzinfo=NULL;
    }
    else 
        *wzinfo=opened;
    return retval;
}

int _load_wzproject(struct wzproject **wzinfo, int fd) {
    struct wzproject *header;
    struct block *blk;
    unsigned int i,j;
    unsigned long anim_first_dword;
    unsigned long links_length;
    int para_size;
    int cur_pos;
    _WZ_DBGVAR
    
    if ((header=(struct wzproject *)malloc(sizeof(struct wzproject)))==NULL)
        return WZ_MEMORY;
    
    // READ header
    {
        lseek(fd,0,SEEK_SET);
        if (READ(fd,header,16)!=16)
            return _WZ_READFILE;
        //printf("there is %d blocks, %d pages\n",header->nblocks,header->npages);
        header->blocks=NULL;
        header->pages=NULL;
        *wzinfo=header;
    };
    
    // READ blocks
    {
        if ((header->blocks=(struct block *)calloc(header->nblocks,sizeof(struct block)))==NULL)
            return WZ_MEMORY;
        lseek(fd,0x30,SEEK_SET);
        for (i=0;i<header->nblocks;i++) {
            blk=&(header->blocks[i]);
            blk->offset=lseek(fd,0x0,SEEK_CUR);
            //blk->nblocks=282;printf("opening block %d (0x%X)\n",i,blk->offset);
            if (READ(fd,blk,12)!=12)
                return _WZ_READFILE;
            blk->blob=blk->name=NULL;
            // Name
            if (blk->flags&WZ_HAS_NAME) {
                if ((blk->name=(char *)calloc(WZ_NAME_SIZE+1,sizeof(char)))==NULL)
                    return WZ_MEMORY;
                if (READ(fd,blk->name,WZ_NAME_SIZE)!=WZ_NAME_SIZE)
                    return _WZ_READFILE;
            };
            // Links
            if (blk->nlinks>WZ_MAX_LINKS)
                return WZ_TOO_MANY_LINKS;
            else if (blk->nlinks>0) {
                links_length=blk->nlinks*sizeof(unsigned long);
                if (READ(fd,&(blk->links[0]),links_length)!=links_length)
                    return _WZ_READFILE;
            };
            // Anim structs
            if (blk->nanims>WZ_MAX_ANIMS)
                return WZ_TOO_MANY_ANIMS;
            for (j=0;j<blk->nanims;j++) {
                if (READ(fd,&anim_first_dword,sizeof(unsigned long))!=sizeof(unsigned long))
                    return _WZ_READFILE;
                if (anim_first_dword==0)
                    continue;
                lseek(fd,-sizeof(unsigned long),SEEK_CUR);
                if (READ(fd,&(blk->anims[i]),sizeof(struct anim))!=sizeof(struct anim))
                    return _WZ_READFILE;
            };
            // Parameters
            /*if ((blk->para=(long int *)calloc((blk->dwords),sizeof(long int)))==NULL)
                return WZ_MEMORY;*/
            para_size=(blk->dwords)*sizeof(long int);
            if (READ(fd,blk->para,para_size)!=para_size)
                return _WZ_READFILE;
            // BLOb
            if (blk->flags&WZ_HAS_BLOB) {
                if (READ(fd,&(blk->blob_size),sizeof blk->blob_size)!=sizeof blk->blob_size)
                    return _WZ_READFILE;
                if ((blk->blob=(char *)malloc(blk->blob_size))==NULL)
                    return WZ_MEMORY;
                if (READ(fd,blk->blob,blk->blob_size)!=(signed)blk->blob_size)
                    return _WZ_READFILE;
                // Align
                cur_pos=lseek(fd,0,SEEK_CUR);
                if (cur_pos%4!=0)
                    lseek(fd,4-cur_pos%4,SEEK_CUR);
            };
        };
    };
    
    // READ pages
    {
        if ((header->pages=(char **)calloc(header->npages,sizeof(char *)))==NULL)
            return WZ_MEMORY;
        for (i=0;i<header->npages;i++) {
            if ((header->pages[i]=(char *)malloc(WZ_NAME_SIZE))==NULL)
                return WZ_MEMORY;
            if (READ(fd,header->pages[i],WZ_NAME_SIZE)<0)
                return _WZ_READFILE;
        };
    };
    
    return 0;
}

/**
 *  TODO: Save WZ project to file
 */

/**
 *  Print wz project dump.
 */
void print_wzproject(struct wzproject *wzinfo) {
    unsigned int i,j;
    struct block *blk;
    char magic[5]="????";
    const char *hrow1="===============================================";
    const char *hrow2="-----------------------------------------------";
    char _byte;
    
    strncpy(magic,(char *)&(wzinfo->magic),4);
    printf("%s\n",hrow1);
    printf("WERKKZEUG DATAFILE DUMP\n");
    printf("%s\n",hrow1);
    printf("\tHeader information:\n");
    printf("Magic:\t\t%s (%d)\n",magic,wzinfo->magic);
    printf("Version:\t%d\n",wzinfo->version);
    printf("Total blocks:\t%d\n",wzinfo->nblocks);
    printf("Total pages:\t%d\n",wzinfo->npages);
    if ((wzinfo->npages>0)&&(wzinfo->pages!=NULL)) {
        printf("%s\n",hrow1);
        printf("\tPages:\n");
        printf("%s\n",hrow2);
        for (i=0;i<wzinfo->npages;i++)
            if (wzinfo->pages[i]!=NULL)
                printf("%d.\t%s\n",i+1,wzinfo->pages[i]);
    };
    if (wzinfo->nblocks>0) {
        printf("%s\n",hrow1);
        printf("\tBlocks:\n");
        for (i=0;i<wzinfo->nblocks;i++) {
            blk=&(wzinfo->blocks[i]);
            printf("%s\n",hrow2);
            printf("BLOCK #%d",i+1);
            if (blk->name!=NULL)
                printf(" (%s)",blk->name);
            printf("\n");
            printf("Offset: 0x%X\n",blk->offset);
            if ((blk->page<wzinfo->npages)&&(wzinfo->pages!=NULL))
                printf("Page:\t\t%s\n",wzinfo->pages[blk->page]);
            else
                printf("Page:\t\tunknown (%d)\n",blk->page);
            printf("Coords:\t\t(%d,%d)\n",blk->xpos,blk->ypos);
            printf("Width:\t\t%d\n",blk->width);
            printf("CID:\t\t$%8.8X\n",blk->cid);
            printf("Links:\t\t%d\n",blk->nlinks);
            for (j=0;j<blk->nlinks;j++)
                printf("\t%d)\t%d\n",j,blk->links[j]);
            printf("Anims:\t\t%d\n",blk->nanims);
            for (j=0;j<blk->nanims;j++);
                //printf("\t%d);
            printf("Para dwords:\t%d\n",blk->dwords);
            for (j=0;j<blk->dwords*4;j++) {
                _byte=((unsigned char *)blk->para)[j];
                if ((j%16==0)&&(j>0))
                    printf("\n");
                printf("%02hhX ",_byte);
            };
            printf("\n");
            printf("Flags:\t\t%d\n",blk->flags);
            if (blk->blob!=NULL) {
                printf("Block has BLOB\n");
                if (blk->cid==0x00030b05)
                    printf("%s",blk->blob);
            };
        };
    };
    printf("%s\n",hrow1);
    printf("\n");
}

void die(char *message) {
    printf("%s\n",message);
    exit(1);
}

void usage(char *program_name) {
    printf("Usage:\t%s wzfile.k1\n",program_name);
}

int main(int argc, char **argv) {
    int fd;
    int status;
    struct wzproject *wzinfo=NULL;
    
    if (argc<2)
        die("Too few arguments, try --help for more info");
    if (!strcmp(argv[1],"--help")) {
        usage(argv[0]);
        return 0;
    }
    
    fd=open(argv[1],O_RDONLY);
    status=load_wzproject(&wzinfo,fd);
    close(fd);
    
    if (status!=0)
        printf("Cannot open k1 file, status code 0x%X\n", status);
    else
        print_wzproject(wzinfo);
    
    delete_wzproject(wzinfo);
    
    return 0;
}
